from django.contrib import admin
from zumanji.models import Project, Revision, Build


class ProjectAdmin(admin.ModelAdmin):
    list_display = ('label',)
    search_fields = ('label',)

admin.site.register(Project, ProjectAdmin)


class RevisionAdmin(admin.ModelAdmin):
    list_display = ('label', 'project')
    list_filter = ('project',)
    search_fields = ('label', 'project__label')

admin.site.register(Revision, RevisionAdmin)


class BuildAdmin(admin.ModelAdmin):
    list_display = ('label', 'project', 'revision', 'num_tests', 'result', 'datetime')
    list_filter = ('datetime', 'project', 'result')
    search_fields = ('label', 'project__label', 'revision__label')
    raw_id_fields = ('revision',)

admin.site.register(Build, BuildAdmin)



import base64
from django.db import models
from django.utils import simplejson


RESULT_CHOICES = tuple((k, k) for k in (
    'success',
    'failed',
    'skipped',
    'deprecated',
))


class GzippedJSONField(models.TextField):
    """
    Slightly different from a JSONField in the sense that the default
    value is a dictionary.
    """
    __metaclass__ = models.SubfieldBase

    def to_python(self, value):
        if isinstance(value, basestring) and value:
            try:
                value = simplejson.loads(base64.b64decode(value).decode('zlib'))
            except Exception:
                return {}
        elif not value:
            return {}
        return value

    def get_prep_value(self, value):
        if value is None:
            return
        return base64.b64encode(simplejson.dumps(value).encode('zlib'))

    def value_to_string(self, obj):
        value = self._get_val_from_obj(obj)
        return self.get_prep_value(value)

    def south_field_triple(self):
        "Returns a suitable description of this field for South."
        from south.modelsinspector import introspector
        field_class = "django.db.models.fields.TextField"
        args, kwargs = introspector(self)
        return (field_class, args, kwargs)


class Project(models.Model):
    label = models.CharField(max_length=64)
    data = GzippedJSONField(default={}, blank=True)

    def __unicode__(self):
        return self.label


class Revision(models.Model):
    project = models.ForeignKey(Project)
    label = models.CharField(max_length=64)
    data = GzippedJSONField(default={}, blank=True)

    class Meta:
        unique_together = (('project', 'label'),)

    def __unicode__(self):
        return self.label


class Build(models.Model):
    project = models.ForeignKey(Project)
    revision = models.ForeignKey(Revision)
    datetime = models.DateTimeField()
    num_tests = models.PositiveIntegerField(default=0)
    total_duration = models.FloatField(default=0.0)
    data = GzippedJSONField(default={}, blank=True)
    result = models.CharField(max_length=16, choices=RESULT_CHOICES, null=True)

    class Meta:
        unique_together = (('revision', 'datetime'),)

    def __unicode__(self):
        return unicode(self.datetime)

    def save(self, *args, **kwargs):
        self.project = self.revision.project
        super(Build, self).save(*args, **kwargs)

    def _get_temporal_sibling(self, datetime_filter, order_field, tag=None):
        filter_args = {
            'project': self.project,
            datetime_filter: self.datetime
        }
        if tag:
            filter_args["tags"] = tag

        try:
            return type(self).objects.filter(
                **filter_args
            ).exclude(
                id=self.id,
            ).order_by(order_field)[0]
        except IndexError:
            return None

    def get_previous_build(self, tag=None):
        return self._get_temporal_sibling('datetime__lt', "-datetime", tag)

    def get_next_build(self, tag=None):
        return self._get_temporal_sibling('datetime__gt', "datetime", tag)


class BuildTag(models.Model):
    builds = models.ManyToManyField(Build, related_name='tags')
    label = models.CharField(max_length=255)

    def __unicode__(self):
        return self.label


class Test(models.Model):
    project = models.ForeignKey(Project)
    revision = models.ForeignKey(Revision)
    build = models.ForeignKey(Build)
    parent = models.ForeignKey('self', null=True)
    label = models.CharField(max_length=255)
    description = models.TextField(null=True)
    num_tests = models.PositiveIntegerField(default=0)
    mean_duration = models.FloatField(default=0.0)
    upper_duration = models.FloatField(default=0.0)
    lower_duration = models.FloatField(default=0.0)
    upper90_duration = models.FloatField(default=0.0)
    data = GzippedJSONField(default={}, blank=True)
    result = models.CharField(max_length=16, choices=RESULT_CHOICES, null=True)

    class Meta:
        unique_together = (('build', 'label'),)

    def __unicode__(self):
        return self.label

    def save(self, *args, **kwargs):
        if self.parent:
            self.build = self.parent.build
        self.revision = self.build.revision
        self.project = self.revision.project
        super(Test, self).save(*args, **kwargs)

    def shortlabel(self):
        if not self.parent:
            return self.label
        return self.label[len(self.parent.label) + 1:]

    def get_test_in_previous_build(self):
        try:
            return type(self).objects.filter(
                build__project=self.project,
                build__datetime__lt=self.build.datetime,
                label=self.label,
            ).exclude(
                build=self.build,
            ).order_by('-build__datetime')[0]
        except IndexError:
            return None

    def get_test_in_next_build(self):
        try:
            return type(self).objects.filter(
                build__project=self.project,
                build__datetime__gt=self.build.datetime,
                label=self.label,
            ).exclude(
                build=self.build,
            ).order_by('build__datetime')[0]
        except IndexError:
            return None

    def get_previous_builds(self):
        return Build.objects.filter(
            project=self.project,
            datetime__lt=self.build.datetime,
            test__label=self.label,
        ).exclude(
            id=self.build.id,
        ).order_by('-datetime')

    def get_context(self):
        # O(N), so dont abuse it
        nodes = []
        parent = self
        while parent:
            nodes.append(parent)
            parent = parent.parent
        nodes.reverse()
        return nodes


class TestData(models.Model):
    project = models.ForeignKey(Project)
    revision = models.ForeignKey(Revision)
    build = models.ForeignKey(Build)
    test = models.ForeignKey(Test)
    key = models.CharField(max_length=32)
    data = GzippedJSONField(default={}, blank=True)

    class Meta:
        unique_together = (('test', 'key'),)

    def save(self, *args, **kwargs):
        self.build = self.test.build
        self.revision = self.build.revision
        self.project = self.revision.project
        super(TestData, self).save(*args, **kwargs)



from django.conf.urls.defaults import *

urlpatterns = patterns('',
    url(r'^$', 'zumanji.views.index', name='index'),
    url(r'^project/(?P<project_label>[^/]+)$', 'zumanji.views.view_project', name='view_project'),
    url(r'^project/(?P<project_label>[^/]+)/upload$', 'zumanji.views.upload_project_build', name='upload_project_build'),
    url(r'^project/(?P<project_label>[^/]+)/tag/(?P<tag_id>\d+)$', 'zumanji.views.view_tag', name='view_tag'),
    url(r'^project/(?P<project_label>[^/]+)/build/(?P<build_id>\d+)$', 'zumanji.views.view_build', name='view_build'),
    url(r'^project/(?P<project_label>[^/]+)/tag/(?P<tag_id>\d+)/build/(?P<build_id>\d+)$', 'zumanji.views.view_build', name='view_build'),
    url(r'^project/(?P<project_label>[^/]+)/build/(?P<build_id>\d+)/report/(?P<test_label>[^/]+)$', 'zumanji.views.view_test', name='view_test'),
)



from django import forms


class UploadJsonForm(forms.Form):
    revision = forms.CharField(max_length=64, required=False,
        help_text="(Optional) Specify a revision for this build. We will try to extract this from the JSON file if not specified.")
    json_file = forms.FileField()



from django.shortcuts import render


class Interface(object):
    template = None

    def __init__(self, data):
        self.data = data

    def render(self):
        if self.template is None:
            raise NotImplementedError

        return render(self.template, {
            'data': self.data,
        })



from django import template
from django.conf import settings
from django.core.urlresolvers import reverse
from django.utils import simplejson
from django.utils.html import mark_safe
from zumanji.helpers import get_historical_data

register = template.Library()


@register.filter
def format_historical(data):
    result = []
    for item in data:
        if item:
            result.append('|'.join(map(str, item)))
        else:
            result.append('')
    return ','.join(result)


@register.filter('range')
def range_filter(value):
    return range(int(value))


@register.filter
def as_json(value):
    return mark_safe(simplejson.dumps(value))


@register.inclusion_tag('zumanji/includes/sparkline.html')
def render_historical_sparkline(test):
    data = get_historical_data(test.build, [test])[test.id]
    historical = []
    for build_id, result in data:
        if build_id:
            historical.append({
                'url': reverse('zumanji:view_test', kwargs={
                    'project_label': test.project.label,
                    'build_id': build_id,
                    'test_label': test.label,
                }),
                'title': 'Build #%s' % build_id,
                'values': result,
            })
        else:
            historical.append({
                'values': result,
            })
    return {
        'columns': [c[1] for c in settings.ZUMANJI_CONFIG['call_types']],
        'test': test,
        'historical': historical,
    }


@register.inclusion_tag('zumanji/includes/test_row.html')
def render_test_row(test):
    data = test.data
    columns = []
    for column, _ in settings.ZUMANJI_CONFIG['call_types']:
        result = data.get(column)
        if not result:
            columns.append(None)
        else:
            columns.append((result.get('mean_duration'), result.get('mean_calls')))

    return {
        'test': test,
        'columns': columns,
    }


@register.inclusion_tag('zumanji/includes/test_columns.html')
def render_test_columns():
    return {
        'columns': [c[1] for c in settings.ZUMANJI_CONFIG['call_types']]
    }


@register.inclusion_tag('zumanji/includes/change_row.html')
def render_change_row(change, compare_build):
    test, data = change

    columns = []
    for column, _ in settings.ZUMANJI_CONFIG['call_types']:
        result = data['interfaces'].get(column)
        if not result:
            columns.append(None)
        else:
            columns.append((result.get('type'), result.get('change')))

    return {
        'test': test,
        'columns': columns,
        'compare_build': compare_build,
    }



import difflib
import itertools
from collections import defaultdict
from django.conf import settings
from django.utils.datastructures import SortedDict
from zumanji.models import Build, Test, TestData

HISTORICAL_POINTS = 25


def get_trace_data(test, previous_test=None):
    if previous_test:
        try:
            previous_trace = previous_test.testdata_set.get(key='trace').data
        except TestData.DoesNotExist:
            previous_trace = []
    else:
        previous_trace = []

    try:
        trace = test.testdata_set.get(key='trace').data
    except TestData.DoesNotExist:
        trace = []

    if not (trace or previous_trace):
        return {}

    previous_trace = SortedDict(('%s_%s' % (x, c['id']), c) for x, c in enumerate(previous_trace))
    trace = SortedDict(('%s_%s' % (x, c['id']), c) for x, c in enumerate(trace))

    seqmatch = difflib.SequenceMatcher()
    seqmatch.set_seqs(previous_trace.keys(), trace.keys())

    trace_diff = (
        {'test': previous_test, 'calls': []},  # left
        {'test': test, 'calls': []},  # left
    )
    for tag, i1, i2, j1, j2 in seqmatch.get_opcodes():
        if tag in ('equal', 'replace'):
            for key in previous_trace.keys()[i1:i2]:
                trace_diff[0]['calls'].append((tag, key, previous_trace[key]))
            for key in trace.keys()[j1:j2]:
                trace_diff[1]['calls'].append((tag, key, trace[key]))
        elif tag == 'delete':
            for key in previous_trace.keys()[i1:i2]:
                trace_diff[0]['calls'].append((tag, key, previous_trace[key]))
                trace_diff[1]['calls'].append((tag, key, None))
        elif tag == 'insert':
            for key in trace.keys()[j1:j2]:
                trace_diff[0]['calls'].append((tag, key, None))
                trace_diff[1]['calls'].append((tag, key, trace[key]))
        else:
            raise ValueError(tag)

    all_calls = dict(previous_trace)
    all_calls.update(trace)

    return {
        'diff': trace_diff,
        'calls': all_calls,
        'num_diffs': sum(sum(1 for t, _, c in n['calls'] if t != 'equal') for n in trace_diff),
    }


def get_historical_data(build, test_list):
    previous_builds = list(Build.objects.filter(
        datetime__lt=build.datetime,
        project=build.project,
    ).exclude(
        id=build.id,
    ).order_by('-datetime')
     .values_list('id', flat=True)[:HISTORICAL_POINTS - 1][::-1])

    previous_tests = list(Test.objects.filter(
        build__in=previous_builds,
        label__in=[t.label for t in test_list]
    )
    .order_by('-build__datetime'))

    historical = defaultdict(lambda: defaultdict(list))
    for test in itertools.chain(previous_tests, test_list):
        history_data = [test.mean_duration]
        for interface, _ in settings.ZUMANJI_CONFIG['call_types']:
            if interface not in test.data:
                history_data.append(0)
            else:
                interface_duration = test.data[interface].get('mean_duration', 0)
                history_data[0] -= interface_duration
                history_data.append(interface_duration)
        historical[test.label][test.build_id] = history_data

    padding = [(None, [])] * HISTORICAL_POINTS
    results = {}
    for test in test_list:
        results[test.id] = (padding + [
            (b, historical[test.label][b])
            for b in (previous_builds + [test.build_id])
        ])[-HISTORICAL_POINTS:]

    return results


def get_changes(previous_build, objects):
    if not (previous_build and objects):
        return {}

    qs = previous_build.test_set.filter(
        label__in=[o.label for o in objects],
    ).select_related('parent')

    previous_build_objects = dict((o.label, o) for o in qs)
    changes = dict()

    # {group: [{notes: notes, type: type}]}
    for obj in objects:
        last_obj = previous_build_objects.get(obj.label)
        obj_changes = {
            'interfaces': {},
            'status': 'new' if last_obj is None else None,
        }
        if last_obj:
            data = obj.data
            last_obj_data = last_obj.data
            for interface, _ in settings.ZUMANJI_CONFIG['call_types']:
                if interface in data:
                    current = data[interface].get('mean_calls', 0)
                else:
                    current = 0

                if interface in last_obj_data:
                    previous = last_obj_data[interface].get('mean_calls', 0)
                else:
                    previous = 0

                change = current - previous
                if change == 0:
                    continue

                obj_changes['interfaces'][interface] = {
                    'current': current,
                    'previous': previous,
                    'change': '+%s' % change if change > 0 else str(change),
                    'type': 'increase' if change > 0 else 'decrease',
                }

        if obj_changes['status'] != 'new' and not obj_changes['interfaces']:
            continue

        changes[obj] = obj_changes

    return sorted(changes.iteritems(),
        key=lambda x: sum(int(i['change']) for i in x[1]['interfaces'].values()),
        reverse=True)



__all__ = ('import_build',)

import datetime
import hashlib
from collections import defaultdict
from zumanji.models import Project, Revision, Build, BuildTag, Test


def convert_timestamp(timestamp):
    return datetime.datetime.strptime(timestamp, '%Y-%m-%dT%H:%M:%S.%f')


def count_leaves_with_tests(labels):
    # test.label: set(leaves)
    leaves = defaultdict(set)
    for label in labels:
        leaves[label].add(label)
        parent = label.rsplit('.', 1)[0]
        leaves[parent].add(label)
        while len(leaves[parent]) > 1 and '.' in parent:
            parent = parent.rsplit('.', 1)[0]
            leaves[parent].add(parent)

    return dict((k, len(v)) for k, v in leaves.iteritems())


def regroup_tests(tests):
    grouped = defaultdict(list)

    for test in tests:
        key = []
        for part in test['id'].split('.')[:-1]:
            key.append(part)
            grouped['.'.join(key)].append(test)

    return sorted(grouped.items(), key=lambda x: x[0])


def with_call_id(data):
    call_id = hashlib.md5(data['interface'])
    call_id.update(data['command'])
    call_id.update(data['filename'])
    call_id.update(data['function'])
    call_id = call_id.hexdigest()

    data['id'] = call_id
    return data


def format_v2_data(data):
    frame = data['stacktrace'][0]

    data['start'] = float(data['start'])
    if data.get('end') is not None:
        data['end'] = float(data['end'])
        duration = data['end'] - data['start']
    else:
        duration = 0.0

    return with_call_id({
        'interface': data['type'],
        'command': data['name'],
        'args': data['args'],
        'function': frame['function'],
        'filename': frame['filename'],
        'lineno': frame['lineno'],
        'duration': duration,
        'time': datetime.datetime.fromtimestamp(data['start']).isoformat(),
        'depth': len(data['stacktrace']),
        'stacktrace': data['stacktrace'],
    })


def format_v1_data(item):
    interface, data = item
    frame = data['stacktrace'][0]

    if interface == 'sql':
        command = data['query']
        args = data['query_params']
    elif interface in ('redis', 'pipelined_redis'):
        command = data['command']
        if 'actions' in data:
            args = data['actions']
        else:
            args = [data['other_args']]
    elif interface == 'cache':
        command = data['action']
        args = [data['key']]
    else:
        command = u':'.join(frame['filename'], frame['function'])
        args = []

    return with_call_id({
        'interface': interface,
        'command': command,
        'args': args,
        'function': frame['function'],
        'filename': frame['filename'],
        'lineno': frame['lineno'],
        'duration': data['duration'],
        'time': data['time'],
        'depth': len(data['stacktrace']),
        'stacktrace': data['stacktrace'],
    })


def percentile(values, percentile=90):
    return values[int(len(values) * (percentile / 100.0))]


def create_test_leaf(build, data, parent, version=1):
    if version == 1:
        format_data = format_v1_data

        interface_data = []
        for interface, values in data.get('interfaces', {}).iteritems():
            for item in values:
                interface_data.append((interface, item))

        interface_data.sort(key=lambda x: x[0])

    elif version == 2:
        format_data = format_v2_data

        interface_data = data.get('calls', [])
        interface_data.sort(key=lambda x: float(x['start']))

    else:
        raise ValueError('version')

    interface_data = [format_data(d) for d in interface_data]

    description = (data.get('doc') or '').strip()

    extra_data = defaultdict(lambda: {
        'mean_calls': 0,
        'mean_duration': 0.0,
    })
    for item in interface_data:
        extra_data[item['interface']]['mean_calls'] += 1
        extra_data[item['interface']]['mean_duration'] += item['duration']

    test = Test.objects.create(
        build=build,
        project=build.project,
        revision=build.revision,
        label=data['id'],
        description=description,
        mean_duration=data['duration'],
        upper90_duration=data['duration'],
        upper_duration=data['duration'],
        lower_duration=data['duration'],
        data=extra_data,
        parent=parent,
    )

    test.testdata_set.create(
        build=test.build,
        revision=test.revision,
        project=test.project,
        key='trace',
        data=interface_data,
    )

    return test


def import_build(data, project=None, revision=None):
    version = int(data.get('version', 1))
    timestamp = convert_timestamp(data['time'])

    project_label = project or data.get('project')
    if not project_label:
        raise ValueError('You must specify a project')

    revision_label = revision or data.get('revision')
    if not revision_label:
        raise ValueError('You must specify a revision')

    # We avoid recreating the core items that might get referenced by an ID in the interface
    project = Project.objects.get_or_create(
        label=project_label,
    )[0]

    revision = Revision.objects.get_or_create(
        project=project,
        label=revision_label,
    )[0]

    build, created = Build.objects.get_or_create(
        project=project,
        revision=revision,
        datetime=timestamp,
    )

    # Clean out old tests
    build.test_set.all().delete()

    # Clean out old tags
    build.tags.all().delete()
    # Add tags
    tag_list = [
        BuildTag.objects.get_or_create(label=tag_name)[0]
        for tag_name in data.get('tags', [])
    ]
    build.tags.add(*tag_list)

    num_tests = 0
    total_duration = 0.0
    tests_by_id = {}
    grouped_tests = regroup_tests(data['tests'])

    # Eliminate useless parents (parents which only have a single child)
    leaf_counts = count_leaves_with_tests((t['id'] for t in data['tests']))

    def find_parent(label):
        if '.' not in label:
            return None

        key = label.split('.')[:-1]
        while key:
            path = '.'.join(key)
            if path in tests_by_id:
                return tests_by_id[path]
            key.pop()
        return None

    grouped_tests = [(l, t) for l, t in grouped_tests if leaf_counts.get(l) >= 1]

    # Create branches
    for label, _ in grouped_tests:
        parent = find_parent(label)

        branch = Test.objects.create(
            parent=parent,
            project=project,
            revision=revision,
            build=build,
            label=label,
        )
        tests_by_id[branch.label] = branch

    for label, tests in reversed(grouped_tests):
        branch = tests_by_id[label]

        # Create any leaves which do not exist yet
        for test_data in (t for t in tests if t['id'] not in tests_by_id):
            test = create_test_leaf(build, test_data, branch, version)
            tests_by_id[test.label] = test

            num_tests += 1
            total_duration += test.mean_duration

        # Update aggregated data
        group_durations = []
        interface_durations = defaultdict(list)

        for test in (tests_by_id[t['id']] for t in tests):
            group_durations.append(test.mean_duration)
            for interface, values in test.data.iteritems():
                interface_durations[interface].append(values)

        group_num_tests = len(group_durations)
        group_total_duration = sum(group_durations)
        group_durations.sort()

        td_data = {}
        for interface, values in interface_durations.iteritems():
            durations = [v['mean_duration'] for v in values]
            td_data[interface] = {
                'mean_calls': sum(v['mean_calls'] for v in values),
                'mean_duration': sum(durations),
                'upper_duration': durations[-1],
                'lower_duration': durations[0],
                'upper90_duration': percentile(durations, 90),
            }

        tests_by_id[branch.label] = Test.objects.filter(id=branch.id).update(
            label=label,
            num_tests=group_num_tests,
            mean_duration=group_total_duration,
            upper_duration=group_durations[-1],
            lower_duration=group_durations[0],
            upper90_duration=percentile(group_durations, 90),
            data=td_data,
        )

    Build.objects.filter(id=build.id).update(
        num_tests=num_tests,
        total_duration=total_duration,
    )

    return build



#!/usr/bin/env python
"""
zumanji
=======

:copyright: (c) 2012 DISQUS
:license: Apache License 2.0, see LICENSE for more details.
"""

from setuptools import setup, find_packages

# Hack to prevent stupid "TypeError: 'NoneType' object is not callable" error
# in multiprocessing/util.py _exit_function when running `python
# setup.py test` (see
# http://www.eby-sarna.com/pipermail/peak/2010-May/003357.html)
try:
    __import__('multiprocessing')
except ImportError:
    pass

tests_require = [
]


install_requires = [
    'django>=1.2',
    'psycopg2',
    'django-crispy-forms>=1.1.4',
    'south',
]

setup(
    name='zumanji',
    version='0.2.2',
    author='DISQUS',
    author_email='opensource@disqus.com',
    url='https://github.com/disqus/zumanji',
    description='A web interface for aggregating results from nose-performance',
    long_description=__doc__,
    package_dir={'': 'src'},
    packages=find_packages('src'),
    zip_safe=False,
    install_requires=install_requires,
    tests_require=tests_require,
    test_suite='runtests.runtests',
    license='Apache License 2.0',
    include_package_data=True,
    classifiers=[
        'Framework :: Django',
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'Operating System :: OS Independent',
        'Topic :: Software Development'
    ],
)



import os.path
from django.db import transaction
from django.utils import simplejson
from django.core.management.base import BaseCommand, CommandError
from optparse import make_option
from zumanji.importer import import_build


class Command(BaseCommand):
    args = '<json_file json_file ...>'
    help = 'Imports the specified JSON files'

    option_list = BaseCommand.option_list + (
        make_option('--project', '-p', dest='project', help='Project Label'),
        make_option('--revision', '-r', dest='revision', help='Revision Label'),
    )

    @transaction.commit_on_success
    def handle(self, *args, **options):
        for json_file in args:

            self.stdout.write('Reading json file %r\n' % (json_file,))

            if not os.path.exists(json_file):
                raise CommandError('Json file %r does not exist' % json_file)

            with open(json_file, 'r') as fp:
                data = simplejson.loads(fp.read())

            build = import_build(data, project=options.get('project'),
                revision=options.get('revision'))
            transaction.commit()

            self.stdout.write('Imported %r (build_id=%r)\n' % (json_file, build.id))



# -*- coding: utf-8 -*-
import datetime
from south.db import db
from south.v2 import SchemaMigration
from django.db import models


class Migration(SchemaMigration):

    def forwards(self, orm):
        # Adding model 'Project'
        db.create_table('zumanji_project', (
            ('id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('label', self.gf('django.db.models.fields.CharField')(max_length=64)),
            ('data', self.gf('django.db.models.fields.TextField')(default={})),
        ))
        db.send_create_signal('zumanji', ['Project'])

        # Adding model 'Revision'
        db.create_table('zumanji_revision', (
            ('id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('project', self.gf('django.db.models.fields.related.ForeignKey')(to=orm['zumanji.Project'])),
            ('label', self.gf('django.db.models.fields.CharField')(max_length=64)),
            ('data', self.gf('django.db.models.fields.TextField')(default={})),
        ))
        db.send_create_signal('zumanji', ['Revision'])

        # Adding unique constraint on 'Revision', fields ['project', 'label']
        db.create_unique('zumanji_revision', ['project_id', 'label'])

        # Adding model 'Build'
        db.create_table('zumanji_build', (
            ('id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('project', self.gf('django.db.models.fields.related.ForeignKey')(to=orm['zumanji.Project'])),
            ('revision', self.gf('django.db.models.fields.related.ForeignKey')(to=orm['zumanji.Revision'])),
            ('datetime', self.gf('django.db.models.fields.DateTimeField')()),
            ('num_tests', self.gf('django.db.models.fields.PositiveIntegerField')(default=0)),
            ('total_duration', self.gf('django.db.models.fields.FloatField')(default=0.0)),
            ('data', self.gf('django.db.models.fields.TextField')(default={})),
            ('result', self.gf('django.db.models.fields.CharField')(max_length=16, null=True)),
        ))
        db.send_create_signal('zumanji', ['Build'])

        # Adding unique constraint on 'Build', fields ['revision', 'datetime']
        db.create_unique('zumanji_build', ['revision_id', 'datetime'])

        # Adding model 'BuildTag'
        db.create_table('zumanji_buildtag', (
            ('id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('label', self.gf('django.db.models.fields.CharField')(max_length=255)),
        ))
        db.send_create_signal('zumanji', ['BuildTag'])

        # Adding M2M table for field builds on 'BuildTag'
        db.create_table('zumanji_buildtag_builds', (
            ('id', models.AutoField(verbose_name='ID', primary_key=True, auto_created=True)),
            ('buildtag', models.ForeignKey(orm['zumanji.buildtag'], null=False)),
            ('build', models.ForeignKey(orm['zumanji.build'], null=False))
        ))
        db.create_unique('zumanji_buildtag_builds', ['buildtag_id', 'build_id'])

        # Adding model 'Test'
        db.create_table('zumanji_test', (
            ('id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('project', self.gf('django.db.models.fields.related.ForeignKey')(to=orm['zumanji.Project'])),
            ('revision', self.gf('django.db.models.fields.related.ForeignKey')(to=orm['zumanji.Revision'])),
            ('build', self.gf('django.db.models.fields.related.ForeignKey')(to=orm['zumanji.Build'])),
            ('parent', self.gf('django.db.models.fields.related.ForeignKey')(to=orm['zumanji.Test'], null=True)),
            ('label', self.gf('django.db.models.fields.CharField')(max_length=255)),
            ('description', self.gf('django.db.models.fields.TextField')(null=True)),
            ('num_tests', self.gf('django.db.models.fields.PositiveIntegerField')(default=0)),
            ('mean_duration', self.gf('django.db.models.fields.FloatField')(default=0.0)),
            ('upper_duration', self.gf('django.db.models.fields.FloatField')(default=0.0)),
            ('lower_duration', self.gf('django.db.models.fields.FloatField')(default=0.0)),
            ('upper90_duration', self.gf('django.db.models.fields.FloatField')(default=0.0)),
            ('data', self.gf('django.db.models.fields.TextField')(default={})),
            ('result', self.gf('django.db.models.fields.CharField')(max_length=16, null=True)),
        ))
        db.send_create_signal('zumanji', ['Test'])

        # Adding unique constraint on 'Test', fields ['build', 'label']
        db.create_unique('zumanji_test', ['build_id', 'label'])

        # Adding model 'TestData'
        db.create_table('zumanji_testdata', (
            ('id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('project', self.gf('django.db.models.fields.related.ForeignKey')(to=orm['zumanji.Project'])),
            ('revision', self.gf('django.db.models.fields.related.ForeignKey')(to=orm['zumanji.Revision'])),
            ('build', self.gf('django.db.models.fields.related.ForeignKey')(to=orm['zumanji.Build'])),
            ('test', self.gf('django.db.models.fields.related.ForeignKey')(to=orm['zumanji.Test'])),
            ('key', self.gf('django.db.models.fields.CharField')(max_length=32)),
            ('data', self.gf('django.db.models.fields.TextField')(default={})),
        ))
        db.send_create_signal('zumanji', ['TestData'])

        # Adding unique constraint on 'TestData', fields ['test', 'key']
        db.create_unique('zumanji_testdata', ['test_id', 'key'])


    def backwards(self, orm):
        # Removing unique constraint on 'TestData', fields ['test', 'key']
        db.delete_unique('zumanji_testdata', ['test_id', 'key'])

        # Removing unique constraint on 'Test', fields ['build', 'label']
        db.delete_unique('zumanji_test', ['build_id', 'label'])

        # Removing unique constraint on 'Build', fields ['revision', 'datetime']
        db.delete_unique('zumanji_build', ['revision_id', 'datetime'])

        # Removing unique constraint on 'Revision', fields ['project', 'label']
        db.delete_unique('zumanji_revision', ['project_id', 'label'])

        # Deleting model 'Project'
        db.delete_table('zumanji_project')

        # Deleting model 'Revision'
        db.delete_table('zumanji_revision')

        # Deleting model 'Build'
        db.delete_table('zumanji_build')

        # Deleting model 'BuildTag'
        db.delete_table('zumanji_buildtag')

        # Removing M2M table for field builds on 'BuildTag'
        db.delete_table('zumanji_buildtag_builds')

        # Deleting model 'Test'
        db.delete_table('zumanji_test')

        # Deleting model 'TestData'
        db.delete_table('zumanji_testdata')


    models = {
        'zumanji.build': {
            'Meta': {'unique_together': "(('revision', 'datetime'),)", 'object_name': 'Build'},
            'data': ('django.db.models.fields.TextField', [], {'default': '{}'}),
            'datetime': ('django.db.models.fields.DateTimeField', [], {}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'num_tests': ('django.db.models.fields.PositiveIntegerField', [], {'default': '0'}),
            'project': ('django.db.models.fields.related.ForeignKey', [], {'to': "orm['zumanji.Project']"}),
            'result': ('django.db.models.fields.CharField', [], {'max_length': '16', 'null': 'True'}),
            'revision': ('django.db.models.fields.related.ForeignKey', [], {'to': "orm['zumanji.Revision']"}),
            'total_duration': ('django.db.models.fields.FloatField', [], {'default': '0.0'})
        },
        'zumanji.buildtag': {
            'Meta': {'object_name': 'BuildTag'},
            'builds': ('django.db.models.fields.related.ManyToManyField', [], {'related_name': "'tags'", 'symmetrical': 'False', 'to': "orm['zumanji.Build']"}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'label': ('django.db.models.fields.CharField', [], {'max_length': '255'})
        },
        'zumanji.project': {
            'Meta': {'object_name': 'Project'},
            'data': ('django.db.models.fields.TextField', [], {'default': '{}'}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'label': ('django.db.models.fields.CharField', [], {'max_length': '64'})
        },
        'zumanji.revision': {
            'Meta': {'unique_together': "(('project', 'label'),)", 'object_name': 'Revision'},
            'data': ('django.db.models.fields.TextField', [], {'default': '{}'}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'label': ('django.db.models.fields.CharField', [], {'max_length': '64'}),
            'project': ('django.db.models.fields.related.ForeignKey', [], {'to': "orm['zumanji.Project']"})
        },
        'zumanji.test': {
            'Meta': {'unique_together': "(('build', 'label'),)", 'object_name': 'Test'},
            'build': ('django.db.models.fields.related.ForeignKey', [], {'to': "orm['zumanji.Build']"}),
            'data': ('django.db.models.fields.TextField', [], {'default': '{}'}),
            'description': ('django.db.models.fields.TextField', [], {'null': 'True'}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'label': ('django.db.models.fields.CharField', [], {'max_length': '255'}),
            'lower_duration': ('django.db.models.fields.FloatField', [], {'default': '0.0'}),
            'mean_duration': ('django.db.models.fields.FloatField', [], {'default': '0.0'}),
            'num_tests': ('django.db.models.fields.PositiveIntegerField', [], {'default': '0'}),
            'parent': ('django.db.models.fields.related.ForeignKey', [], {'to': "orm['zumanji.Test']", 'null': 'True'}),
            'project': ('django.db.models.fields.related.ForeignKey', [], {'to': "orm['zumanji.Project']"}),
            'result': ('django.db.models.fields.CharField', [], {'max_length': '16', 'null': 'True'}),
            'revision': ('django.db.models.fields.related.ForeignKey', [], {'to': "orm['zumanji.Revision']"}),
            'upper90_duration': ('django.db.models.fields.FloatField', [], {'default': '0.0'}),
            'upper_duration': ('django.db.models.fields.FloatField', [], {'default': '0.0'})
        },
        'zumanji.testdata': {
            'Meta': {'unique_together': "(('test', 'key'),)", 'object_name': 'TestData'},
            'build': ('django.db.models.fields.related.ForeignKey', [], {'to': "orm['zumanji.Build']"}),
            'data': ('django.db.models.fields.TextField', [], {'default': '{}'}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'key': ('django.db.models.fields.CharField', [], {'max_length': '32'}),
            'project': ('django.db.models.fields.related.ForeignKey', [], {'to': "orm['zumanji.Project']"}),
            'revision': ('django.db.models.fields.related.ForeignKey', [], {'to': "orm['zumanji.Revision']"}),
            'test': ('django.db.models.fields.related.ForeignKey', [], {'to': "orm['zumanji.Test']"})
        }
    }

    complete_apps = ['zumanji']


from django.core.urlresolvers import reverse
from django.http import HttpResponseRedirect
from django.shortcuts import render, get_object_or_404
from django.utils import simplejson
from zumanji.forms import UploadJsonForm
from zumanji.helpers import get_trace_data, get_changes
from zumanji.models import Project, Build, BuildTag, Test
from zumanji.importer import import_build


def index(request):
    build_list = list(Build.objects
        .order_by('-datetime')
        .select_related('revision', 'project'))

    return render(request, 'zumanji/index.html', {
        'build_list': build_list,
    })


def view_project(request, project_label):
    project = get_object_or_404(Project, label=project_label)

    build_list = list(Build.objects
        .filter(project=project)
        .order_by('-datetime')
        .select_related('revision', 'project'))

    return render(request, 'zumanji/project.html', {
        'project': project,
        'build_list': build_list,
    })


def view_tag(request, project_label, tag_id):
    project = get_object_or_404(Project, label=project_label)
    tag = get_object_or_404(BuildTag, pk=tag_id)

    build_list = list(Build.objects
        .filter(project=project, tags=tag)
        .order_by('-datetime')
        .select_related('revision', 'project'))

    return render(request, 'zumanji/tag.html', {
        'project': project,
        'tag': tag,
        'build_list': build_list,
    })


def view_build(request, project_label, build_id, tag_id=None):
    filter_args = dict(project__label=project_label, id=build_id)
    tag = None
    if tag_id:
        tag = get_object_or_404(BuildTag, id=tag_id)
        filter_args["tags"] = tag

    build = get_object_or_404(Build, **filter_args)
    project = build.project
    previous_build = build.get_previous_build(tag=tag)
    next_build = build.get_next_build(tag=tag)

    test_list = list(build.test_set
        .filter(parent__isnull=True)
        .order_by('-upper90_duration'))

    compare_with = request.GET.get('compare_with')
    if compare_with:
        try:
            compare_build = Build.objects.get(project__label=project_label, id=compare_with)
        except Build.DoesNotExist:
            compare_build = None
    else:
        compare_build = previous_build

    changes = get_changes(compare_build, test_list)

    return render(request, 'zumanji/build.html', {
        'project': project,
        'tag': tag,
        'build': build,
        'previous_build': previous_build,
        'compare_build': compare_build,
        'next_build': next_build,
        'test_list': test_list,
        'changes': changes,
    })


def view_test(request, project_label, build_id, test_label):
    test = get_object_or_404(Test, project__label=project_label, build=build_id, label=test_label)
    project = test.project
    build = test.build

    test_list = list(Test.objects.filter(parent=test)
        .order_by('-upper90_duration')
        .select_related('parent'))

    # this is actually a <Test>
    previous_test_by_build = test.get_test_in_previous_build()
    next_test_by_build = test.get_test_in_next_build()

    breadcrumbs = [
        (reverse('zumanji:view_build', kwargs={'project_label': project.label, 'build_id': build.id}), 'Build #%s' % build.id)
    ]
    last = ''
    for node in test.get_context():
        node_label = node.label[len(last):]
        breadcrumbs.append(
            (reverse('zumanji:view_test', kwargs={
                'project_label': project.label,
                'build_id': build.id,
                'test_label': node.label,
            }), node_label)
        )
        last = node.label + '.'  # include the dot

    previous_builds = list(test.get_previous_builds().select_related('revision')[:50])

    compare_with = request.GET.get('compare_with')
    if compare_with:
        try:
            compare_build = Build.objects.get(project__label=project_label, id=compare_with)
        except Build.DoesNotExist:
            compare_build = None
    else:
        compare_build = previous_test_by_build.build if previous_test_by_build else None

    if compare_build:
        try:
            compare_test = compare_build.test_set.get(label=test.label)
        except Test.DoesNotExist:
            compare_test = None
    else:
        compare_test = None

    trace_results = get_trace_data(test, compare_test)
    if previous_test_by_build:
        tests_to_check = test_list
        changes = get_changes(compare_build, tests_to_check)
    else:
        changes = []

    return render(request, 'zumanji/test.html', {
        'breadcrumbs': breadcrumbs,
        'project': project,
        'build': build,
        'previous_test_by_build': previous_test_by_build,
        'next_test_by_build': next_test_by_build,
        'previous_builds': previous_builds,
        'test': test,
        'test_list': test_list,
        'changes': changes,
        'compare_build': compare_build,
        'trace_results': trace_results,
    })


def upload_project_build(request, project_label):
    project = get_object_or_404(Project, label=project_label)

    form = UploadJsonForm(request.POST or None, request.FILES or None)
    if form.is_valid():
        data = simplejson.loads(request.FILES['json_file'].read())

        try:
            build = import_build(data, project=project.label, revision=form.cleaned_data.get('revision'))
        except Exception, e:
            form.errors['json_file'] = unicode(e)
        else:
            return HttpResponseRedirect(reverse('zumanji:view_build', kwargs={
                'project_label': project.label, 'build_id': build.id}))

    return render(request, 'zumanji/upload_build.html', {
        'project': project,
        'form': form,
    })
